//
//  AppDelegate.m
//  PeerFilter
//
//  Created by U Kong Wong on 7/21/12.
//  Copyright 2012 Eric U Kong Wong. All rights reserved.
//

#import <ServiceManagement/ServiceManagement.h>
#import <Security/Authorization.h>
#import <sys/socket.h>
#import <sys/un.h>
#import "AppDelegate.h"
#import "KextControl.h"
#import "ListManager.h"

static NSString *helperJobLabel = @"com.github.mtrbean.PeerFilter.PeerFilterHelper";

@implementation AppDelegate

@synthesize window = _window;
@synthesize filterStatus = _filterStatus;
@synthesize logWindow = _logWindow;
@synthesize listView = _listView;
@synthesize kextControl = _kextControl;
@synthesize listManager = _listManager;


- (void)appendLog:(NSString *)str {
    NSTextStorage *textStorage = [self.logWindow textStorage];
    [[textStorage mutableString] appendString:str];
    [self.logWindow setNeedsDisplay:YES];
}
	
- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    NSError *error = nil;
    NSBundle *mainBundle = [NSBundle mainBundle];
    NSURL *kextPath = [mainBundle URLForResource:@"PeerFilter" withExtension:@"kext"];
    NSURL *TargetPath = [NSURL URLWithString:[NSString stringWithUTF8String:KEXT_PATH]];
    [[NSFileManager defaultManager] fileExistsAtPath:[NSString stringWithUTF8String:KEXT_PATH]];
    [[NSFileManager defaultManager] copyItemAtURL:kextPath toURL:TargetPath error:&error];
    
    // find helper
    CFDictionaryRef jobSpec = SMJobCopyDictionary(kSMDomainSystemLaunchd, (CFStringRef)helperJobLabel);
    
    if (!jobSpec) {
        // bless and install helper
       
        if (![self installHelperWithLabel:helperJobLabel error:&error]) {
            NSLog(@"Error: %@", error);
            [error release];
        } else {
            [self appendLog:@"PeerFilterHelper installed.\n"];   
        }
        CFRelease(jobSpec);
    }
    
    // start helper in a separate thread    
    opQueue = [[NSOperationQueue alloc] init];
    
    [opQueue addOperationWithBlock:^(void) {
        BOOL helperLaunched = [self launchHelperWithSocketPath:@"/var/tmp/peerfilterhelper.sock"];
        if (helperLaunched) {
            [self performSelectorOnMainThread:@selector(appendLog:) 
                                   withObject:@"PeerFilterHelper launched\n" 
                                waitUntilDone:YES];
        }
        
        // connect to kext
        KextControl *ctl = [[KextControl alloc] init];
        NSString *logMsg;
        if ([ctl connect]) {
            logMsg = @"Connected to Kext\n";
        }
        else {
            logMsg = @"Failed to connect to Kext\n";
        }
        [self performSelectorOnMainThread:@selector(appendLog:) 
                               withObject:logMsg 
                            waitUntilDone:YES];
        
        // set up input stream from kext to logWindow
        [ctl.inputStream setDelegate:self];
        [ctl.inputStream scheduleInRunLoop:[NSRunLoop mainRunLoop]
                                                forMode:NSDefaultRunLoopMode];
        [ctl.inputStream open];
        
        self.kextControl = ctl; 
        [ctl release];       
    }];
    
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
    [opQueue cancelAllOperations];
    [opQueue waitUntilAllOperationsAreFinished];
    [opQueue release];
    [self.kextControl.inputStream close];
    [self.kextControl.inputStream removeFromRunLoop:[NSRunLoop mainRunLoop]
                                            forMode:NSDefaultRunLoopMode];
    [self.kextControl.inputStream setDelegate:nil];
    if  ([self.kextControl active])
        [self.kextControl stop];
    self.kextControl = nil;
    
    return NSTerminateNow;
}

- (BOOL)applicationShouldHandleReopen:(NSApplication *)sender hasVisibleWindows:(BOOL)flag
{
    if (flag) {
        return NO;
    }
    else {
        [self.window orderFront:self];
        return YES;
    }
}

- (void)stream:(NSStream *)stream handleEvent:(NSStreamEvent)eventCode
{
    switch (eventCode) {
        case NSStreamEventOpenCompleted:
            [self appendLog:@"Communication stream with packet filter opened.\n"];
            break;
            
        case NSStreamEventHasBytesAvailable:
        {
            uint8_t buf[1024];
            
            NSInteger actuallyRead = [(NSInputStream *)stream read:buf maxLength:sizeof(buf)];
            if (actuallyRead > 0) {
                NSString *output = [[NSString alloc] initWithBytes:(const void*)buf length:actuallyRead encoding:NSASCIIStringEncoding];
                [self appendLog:output];
                [output release];
            }
            else {
                NSLog(@"No Buffer!");
            }
            break;
        }
        case NSStreamEventEndEncountered:
        {
            [stream close];
            [self appendLog:@"Communication stream with packet filter closed.\n"];
            break;
        }
        default:
            break;
    }
}

- (BOOL)installHelperWithLabel:(NSString *)label error:(NSError **)error
{
    BOOL result = NO;
    
    AuthorizationItem authItem      = { kSMRightBlessPrivilegedHelper, 0, NULL, 0 };
    AuthorizationRights authRights  = { 1, &authItem };
    AuthorizationFlags flags        =   kAuthorizationFlagDefaults              | 
                                        kAuthorizationFlagInteractionAllowed    |
                                        kAuthorizationFlagPreAuthorize          |
                                        kAuthorizationFlagExtendRights;
    
    AuthorizationRef authRef = NULL;
    
    /* Obtain the right to install privileged helper tools (kSMRightBlessPrivilegedHelper). */
    OSStatus status = AuthorizationCreate(&authRights, kAuthorizationEmptyEnvironment, flags, &authRef);
    if (status != errAuthorizationSuccess) {
        NSLog(@"Failed to create AuthorizationRef, return code %i", status);
    } else {
        /* This does all the work of verifying the helper tool against the application
         * and vice-versa. Once verification has passed, the embedded launchd.plist
         * is extracted and placed in /Library/LaunchDaemons and then loaded. The
         * executable is placed in /Library/PrivilegedHelperTools.
         */
        result = SMJobBless(kSMDomainSystemLaunchd, (CFStringRef)label, authRef, (CFErrorRef *)error);
        AuthorizationFree(authRef, kAuthorizationFlagDefaults);
    }
    
    return result;
}

- (BOOL)launchHelperWithSocketPath:(NSString *)unixSocket
{
    int result = -1;
    int sfd = socket(PF_UNIX, SOCK_STREAM, 0);
    if (sfd != -1){
        
        struct sockaddr_un saddr;
        (void)bzero(&saddr, sizeof(saddr));
        saddr.sun_family = AF_UNIX;
        const char* path = [unixSocket UTF8String];
        strncpy(saddr.sun_path, path, strlen(path));
        
        // connect to the designated socket will cause launchd to launch the helper
        result = connect(sfd, (struct sockaddr *)&saddr, sizeof(saddr));
        
        // wait for the complete signal from the helper, with 1 sec timeout
        char buf[10];
        struct timeval tv;
        tv.tv_sec = 1;  // 1 sec timeout
        tv.tv_usec = 0;  // Not init'ing this can cause strange errors
        setsockopt(sfd, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv,sizeof(tv));
        recv(sfd, buf, sizeof(buf), 0);
        
        close(sfd);
    }
    
	return (result == 0);
}

- (IBAction)toggleFilter:(id)sender {
    int ret;
    
    if ([self.kextControl active]) {
        
        ret = [self.kextControl stop];
        if (ret==0) {
            [self.filterStatus setLabel:@"Off"];
        }
        
    }
    else {

        if (self.listManager.dirty) {
            struct pf_iprange_list p = [self.listManager optimizedActiveLists];
            [self.kextControl setListTo:p];
            NSString *log = [NSString stringWithFormat:@"Block list(s) loaded: %d entries\n", p.len];
            [self appendLog:log];
        }
        ret = [self.kextControl start];
        if (ret==0) {
            [self.filterStatus setLabel:@"On"];
        }
        
    }
}

- (IBAction)addList:(id)sender {
    NSOpenPanel *openDialog = [NSOpenPanel openPanel];
    [openDialog setCanChooseFiles:YES];
    [openDialog setAllowsMultipleSelection:YES];
    [openDialog setCanChooseDirectories:NO];
    [openDialog setPrompt:@"Select"];
    if ([openDialog runModal] == NSFileHandlingPanelOKButton) {
        for (NSURL* url in [openDialog URLs]) {
            [self.listManager addListFromURL:url];
        }
    }
    [self.listView reloadData];
}

- (IBAction)removeList:(id)sender {
    NSIndexSet *selectedIndexes = [self.listView selectedRowIndexes];
    [self.listManager removeListsAtIndexes:selectedIndexes];
    [self.listView reloadData];
}
@end
