//
//  ListManager.m
//  PeerFilter
//
//  Created by U Kong Wong on 7/27/12.
//  Copyright (c) 2012 Eric U Kong Wong. All rights reserved.
//

#import "ListManager.h"
#import "GzipInputStream.h"
#include <vector>
#include <algorithm>
#include <zlib.h>
#include "PeerFilterIpRange.h"

#include <arpa/inet.h>

@interface ListManager() {
@private
    NSMutableArray *_listNames;      // for display in tableView
    NSMutableArray *_listStates;     // is active (YES) or inactive (NO)    
    NSMutableDictionary *_listsMap;  // NSString to NSURL mapping
    
    std::vector<struct pf_iprange> _activeRanges; // actual ip range entries
}
@end

@implementation ListManager

@synthesize dirty;

- (void)dealloc
{
    [_listNames release];
    [_listStates release];
    [_listsMap release];
    [super dealloc];
}

- (id)init
{
    self = [super init];
    _listNames = [[NSMutableArray alloc] init];
    _listStates = [[NSMutableArray alloc] init];
    _listsMap = [[NSMutableDictionary alloc] init];  
    self.dirty = NO;
    return self;
}

- (NSInteger)numberOfRowsInTableView:(NSTableView *)tableView
{
    if (_listNames) {
        return [_listNames count];
    }
    return 0;
}

- (id)tableView:(NSTableView *)tableView objectValueForTableColumn:(NSTableColumn *)tableColumn row:(NSInteger)row
{
    return [_listStates objectAtIndex:row];
}

- (void)tableView:(NSTableView *)tableView setObjectValue:(id)value forTableColumn:(NSTableColumn *)tableColumn row:(NSInteger)row
{
    [_listStates replaceObjectAtIndex:row withObject:value];
    self.dirty = YES;
}

- (NSCell*)tableView:(NSTableView *)tableView dataCellForTableColumn:(NSTableColumn *)tableColumn row:(NSInteger)row
{
    NSCell *cell = [tableColumn dataCellForRow:row];
    [cell setTitle:[_listNames objectAtIndex:row]];
    [cell setState:[[_listStates objectAtIndex:row] integerValue]];
    return cell;
}

- (void)addListFromURL:(NSURL *)URL
{
    NSString *key = [URL lastPathComponent];
    [_listNames addObject:key];
    [_listStates addObject:[NSNumber numberWithBool:YES]];
    [_listsMap setObject:URL forKey:key];
    self.dirty = YES;
}

- (void)removeListsAtIndexes:(NSIndexSet *)indexes
{
    NSArray *listsToBeRemoved = [_listNames objectsAtIndexes:indexes];
    [_listNames removeObjectsAtIndexes:indexes];
    [_listStates removeObjectsAtIndexes:indexes];
    [_listsMap removeObjectsForKeys:listsToBeRemoved];
    self.dirty = YES;
}

#pragma mark pf_iprange comparison functions
inline bool less_than(struct pf_iprange const& a, struct pf_iprange const& b) 
{
    return a.start < b.start;
}

inline bool adjacent(struct pf_iprange& a, struct pf_iprange& b) 
{
    if (a.start <= b.start && b.start <= a.end) {
        a.end = std::min(a.end, b.end);
        return true;
    }
    else {
        return false;
    }
}


- (struct pf_iprange_list)optimizedActiveLists
{
    _activeRanges.clear();
    
    for (NSUInteger i = 0; i < [_listNames count]; i++) {
        if ([[_listStates objectAtIndex:i] boolValue]) {
            NSURL *url = [_listsMap objectForKey:[_listNames objectAtIndex:i]];
            GzipInputStream *gis = [[GzipInputStream alloc] initWithURL:url];
            [gis open];
            NSString *line;
            while ((line = [gis readLine])) {
                @try {
                    pf_iprange ipr = [self parseLine:line];
                    _activeRanges.push_back(ipr);
                }
                @catch (NSException *exception) {
                    continue;
                }
            }
            [gis close];
            [gis release];
        }
    }
    
    std::sort(_activeRanges.begin(), _activeRanges.end(), less_than);
    
    std::vector<pf_iprange>::iterator it;
    it = std::unique(_activeRanges.begin(), _activeRanges.end(), adjacent);
    _activeRanges.resize(it - _activeRanges.begin());
    
    pf_iprange_list p;
    p.len = (uint32_t)_activeRanges.size();
    p.ranges = &_activeRanges[0];
    self.dirty = NO;
    
    return p;
}

inline uint32_t ip_to_uint(char const* ip) {
    struct in_addr sa;
    if (inet_pton(AF_INET, ip, &sa) > 0)
        return ntohl(sa.s_addr);
    else
        return 0;
}

- (struct pf_iprange)parseLine:(NSString *)line
{
    struct pf_iprange current;
    NSRange r = [line rangeOfString:@":" options:NSBackwardsSearch];
    if (r.location != NSNotFound) {
        NSArray * endpoints = [[line substringFromIndex:r.location+1] componentsSeparatedByString:@"-"];
        if ([endpoints count]==2) {
            current.start = ip_to_uint([[endpoints objectAtIndex:0] UTF8String]);
            current.end = ip_to_uint([[endpoints objectAtIndex:1] UTF8String]);
            if (current.start && current.end) {
                if (current.start > current.end)
                    std::swap(current.start, current.end);
                return current;
            }
        }
    }
    @throw [NSException exceptionWithName:@"Invalid line" 
                                   reason:@"Unknown format" 
                                 userInfo:nil];
}

@end
