/*
 *  PeerFilterKext.c
 *
 *  Copyright 2012 Eric U Kong Wong. All rights reserved.
 *
 *  This file is part of PeerFilter.
 *
 * PeerFilter is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * PeerFilter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerFilter.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <mach/mach_types.h>
#include <libkern/libkern.h>
#include <libkern/OSKextLib.h>
#include <sys/kernel_types.h> 
#include <sys/systm.h> 
#include <sys/malloc.h> 
#include <sys/kpi_mbuf.h> 
#include <sys/kern_control.h>
#include <netinet/ip.h> 
#include <netinet/kpi_ipfilter.h>
#include "PeerFilterKernControl.h"
#include "PeerFilterIpRange.h"

#pragma mark static global
static struct pf_stats g_filter_stats;

static ipfilter_t g_filter_ref;
static kern_ctl_ref g_ctl_ref;

static boolean_t g_filter_registered = FALSE; 
static boolean_t g_ctl_registered = FALSE; 
static boolean_t g_filter_detached = TRUE; 
static struct {
    boolean_t connected;
    u_int32_t unit;
} g_ctl_client_info;
static struct pf_iprange_list g_pf_list;

#pragma mark forward declarations
kern_return_t PeerFilterKext_start (kmod_info_t* ki, void* d);
kern_return_t PeerFilterKext_stop (kmod_info_t* ki, void* d);
static int pf_filter_load();
static int pf_filter_unload();
static struct kern_ctl_reg g_kern_ctl;

#pragma mark logging
static void log(const char* fmt, ...) {
    va_list args;
    va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);
    if (g_ctl_client_info.connected) {
        // todo handle string longer than 1024 bytes
        char buf[1024];
        va_start(args, fmt);
        int len = vsnprintf(buf, sizeof(buf), fmt, args);
        va_end(args);
        if (len>0) {
            /*  
                The CTL_DATA_EOR flag can be used for the enqueue data and enqueue mbuf
                functions to mark the end of a record. Since this is a SOCK_DGRAM socket,
                specifying 0 will cause the current data to be appended to the previous
                one in the socket without a message boundary.
             */
            int retval = ctl_enqueuedata(g_ctl_ref, g_ctl_client_info.unit, buf, len+1, CTL_DATA_EOR);
            if (retval)
                printf("ctl_enqueuedata failed %d\n", retval);
        }
    }
}

#pragma mark KPI_control
static void free_iprange_list() {
    if (g_pf_list.ranges) {
        FREE(g_pf_list.ranges, 0);
        g_pf_list.ranges = NULL;
        g_pf_list.len = 0;
    }
}

static errno_t pf_filter_ctl_connect(kern_ctl_ref ctl_ref, 
                                     struct sockaddr_ctl *sac, 
                                     void** unitinfo) {
    // only one connection is allowed
    if (!g_ctl_client_info.connected) {
        OSKextRetainKextWithLoadTag(OSKextGetCurrentLoadTag()); // turn on autounload
        log("PeerFilter client with pid=%d connected\n", proc_selfpid());
        g_ctl_client_info.connected = TRUE;
        g_ctl_client_info.unit = sac->sc_unit;
        return 0;
    }
    return EPERM;
}

static errno_t pf_filter_ctl_disconnect(kern_ctl_ref ctl_ref, 
                                        u_int32_t unit, 
                                        void* unitinfo) {
    pf_filter_unload();
    // release for next autounload scan
    OSKextReleaseKextWithLoadTag(OSKextGetCurrentLoadTag());
    g_ctl_client_info.connected = FALSE;
    g_ctl_client_info.unit = 0;
    log("PeerFilter client with pid=%d disconnected\n", proc_selfpid());
    return 0;
}

// handle getsockopt call from client
static errno_t pf_filter_ctl_get(kern_ctl_ref ctl_ref, 
                                 u_int32_t unit, 
                                 void *unitinfo, 
                                 int opt,
                                 void *data, 
                                 size_t *len) {
    int ret = 0;
    switch (opt) {
        case PF_CONTROL_STAT:
            if (*len < sizeof(g_filter_stats)) {
                ret = ENOBUFS;
            }
            else {
                *len = sizeof(g_filter_stats);
                memcpy(data, &g_filter_stats, *len);
            }
            break;
            
        case PF_CONTROL_ISACTIVE:
            if (*len < sizeof(g_filter_registered)) {
                ret = ENOBUFS;
            }
            else {
                memcpy(data, &g_filter_registered, *len);
            }
            break;
            
        default:
            ret = ENOTSUP;
            break;
    }
    return ret;
}

// handle setsockopt call from client
static errno_t pf_filter_ctl_set(kern_ctl_ref ctl_ref, 
                                 u_int32_t unit, 
                                 void* unitinfo, 
                                 int opt,
                                 void* data, 
                                 size_t len) {
    int ret = 0;
    switch (opt) {
        case PF_CONTROL_RESET:
            log("PeerFilter: PF_CONTROL_RESET\n");
            if (!g_filter_detached) {
                ret = EBUSY;
                break;
            }
            free_iprange_list();
            log("PeerFilter: list reset\n");
            break;
            
        case PF_CONTROL_SET_LIST:
            log("PeerFilter: PF_CONTROL_SET_LIST\n");
            if (data == NULL || len!=sizeof(struct pf_iprange_list)) {
                ret = EINVAL;
                log("PeerFilter: Invalid list size");
                break;
            }
            if (!g_filter_detached) {
                ret = EBUSY;
                log("PeerFilter: List cannot be changed while filter is active");
                break;
            }
            free_iprange_list();
            struct pf_iprange_list *user_list = (struct pf_iprange_list *)data;
            size_t size = user_list->len * sizeof(struct pf_iprange);
            if (size == 0) {
                ret = EINVAL;
                log("Empty List...");
                break;
            }
            MALLOC(g_pf_list.ranges, 
                   struct pf_iprange*,
                   size,
                   0,
                   M_WAITOK);
            if (g_pf_list.ranges == NULL) {
                ret = ENOMEM;
                log("PeerFilter: Memory allocation failed\n");
                break;
            }
            ret = copyin((user_addr_t)user_list->ranges, g_pf_list.ranges, size);
            if (ret) {
                log("PeerFilter: copyin() returned %d\n", ret);
                free_iprange_list();
                break;
            }
            g_pf_list.len = user_list->len; 
            log("PeerFilter: %lu range(s) installed\n", user_list->len);
            break;
            
        case PF_CONTROL_START:
            log("PeerFilter: PF_CONTROL_START\n");
            ret = pf_filter_load();
            break;
            
        case PF_CONTROL_STOP:
            log("PeerFilter: PF_CONTROL_STOP\n");
            ret = pf_filter_unload();
            break;
            
        default:
            log("PeerFilter: Unknown Operation");
            ret = ENOTSUP;
            break;
    }
    return ret;
}

#pragma mark ipfilter
static int check_ip_packet(mbuf_t* data, int dir) {
    
    int action = 0; // default action is normal process
    struct ip *ip_header = (struct ip*)mbuf_data(*data); 
    
    if (ip_header == NULL)
        goto bail;
    
    if (g_pf_list.ranges == NULL)
        goto bail;
    
    if (ip_header->ip_v != 4) 
        goto bail;
    
    struct in_addr sa;
    if (dir==kPFiltDirOut)
        sa = ip_header->ip_dst;
    else if (dir==kPFiltDirIn)
        sa = ip_header->ip_src;
    uint32_t ip = ntohl(sa.s_addr);
    
    struct pf_iprange const* p = pf_inranges(&g_pf_list, ip);
    
    if (p) {
        // block
        action = 1;
        log("packet blocked: %d\n", ip);
    
        switch (ip_header->ip_p) {
            case IPPROTO_TCP:
                g_filter_stats.tcp_packets[dir]++;
                break;
            case IPPROTO_UDP:
                g_filter_stats.udp_packets[dir]++;
                break;
            case IPPROTO_ICMP:
                g_filter_stats.icmp_packets[dir]++;
                break;
            default:
                g_filter_stats.other_packets[dir]++;
                break;
        }
    }
    
    return action;
    
bail:
    return 0; // process normally
}

static errno_t pf_filter_output(void* cookie, mbuf_t* data, ipf_pktopts_t options) { 
    if (data) {
        return check_ip_packet(data, kPFiltDirOut); 
    }
    return 0; 
} 

static errno_t pf_filter_input(void* cookie, mbuf_t* data, int offset, u_int8_t protocol) { 
    if (data) {
        return check_ip_packet(data, kPFiltDirIn); 
    }
    return 0; 
} 

static void pf_filter_detach(void* cookie) { 
    /* cookie isn't dynamically allocated, no need to free in this case */
    struct pf_stats* stats =  (struct pf_stats*) cookie;
    log("UDP_IN: %lu\nUDP_OUT: %lu\n",
           stats->udp_packets[kPFiltDirIn],
           stats->udp_packets[kPFiltDirOut]);
    log("TCP_IN: %lu\nTCP_OUT: %lu\n",
           stats->tcp_packets[kPFiltDirIn],
           stats->tcp_packets[kPFiltDirOut]);
    log("ICMP_IN: %lu\nICMP_OUT: %lu\n",
           stats->icmp_packets[kPFiltDirIn],
           stats->icmp_packets[kPFiltDirOut]);
    log("OTHER_IN: %lu\nOTHER_OUT: %lu\n",
           stats->other_packets[kPFiltDirIn],
           stats->other_packets[kPFiltDirOut]);
    log("PeerFilter Detached\n");
    g_filter_detached = TRUE; 
}  

static struct ipf_filter g_pf_filter = {  
    &g_filter_stats, 
    BUNDLE_ID, 
    pf_filter_input, 
    pf_filter_output, 
    pf_filter_detach 
};

static struct kern_ctl_reg g_kern_ctl = {
    BUNDLE_ID,
    0, //ctl_id
    0, //ctl_unit
    0, //ctl_flags
    0, //ctl_sendsize
    0, //ctl_recvsize
    pf_filter_ctl_connect,
    pf_filter_ctl_disconnect,
    NULL,
    pf_filter_ctl_set,
    pf_filter_ctl_get
};
    
static errno_t pf_filter_load() {
    if (g_filter_registered)
        return EBUSY; // already loaded, please unload first

    if (g_pf_list.ranges == NULL) {
        log("Filter list not set\n");
        return EPERM; // please set filters first
    }
    
    int result = ipf_addv4(&g_pf_filter, &g_filter_ref);
    if (result == 0) {
        g_filter_registered = TRUE;
        log("PeerFilter Activated\n");
    }
    return result;
}
    
static errno_t pf_filter_unload() {
    int result = 0;
    
    if (g_filter_registered) {
        result = ipf_remove(g_filter_ref);
        if (result == 0) {
            g_filter_registered = FALSE;
            log("PeerFilter Deactivated\n");
        }
        else {
            log("PeerFilter Cannot be deactivated\n");
        }
    }
    
    /* We need to ensure filter is detached before we return */
    if (!g_filter_detached)
        result = EBUSY; // Try unloading again.
       
    log("PeerFilter: pf_filter_unload returned %d\n", result);
    return result;
}

#pragma mark kext_entry
kern_return_t PeerFilterKext_start (kmod_info_t* ki, void* d) {
    g_ctl_client_info.connected = FALSE;
    g_ctl_client_info.unit = 0;
    g_pf_list.len = 0;
    g_pf_list.ranges = NULL;
    
    int result = ctl_register(&g_kern_ctl, &g_ctl_ref);
    if (result)
        goto bail;

    g_ctl_registered = TRUE;
    log("PeerFilter KEXT Loaded. ID-UNIT:%d-%d\n", 
           g_kern_ctl.ctl_id, 
           g_kern_ctl.ctl_unit);
    
    return KERN_SUCCESS;
    
bail:
    return KERN_FAILURE;
}


kern_return_t PeerFilterKext_stop (kmod_info_t* ki, void* d) {
    int result = pf_filter_unload();
    if (result)
        goto bail;

    if (g_ctl_registered)
        result = ctl_deregister(g_ctl_ref);
    if (result)        
        goto bail;

    g_ctl_registered = FALSE;
    log("PeerFilter KEXT Unloaded\n");
    
    free_iprange_list();
    return KERN_SUCCESS;
    
bail:
    free_iprange_list();
    return KERN_NO_ACCESS; // Try unloading again.
}
